/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022                        ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#include <pecunia/Codes.h>
using pecunia::currency::Iso4217Codes;
#include <pecunia/Money.h>
using pecunia::currency::Money;
#include <QDate>
using ::QDate;
#include <QObject>
using ::QObject;
// using Q_OBJECT
#include <QTest>
// using QTEST_MAIN
// using QVERIFY_EXCEPTION_THROWN
#include <qt-mocks/Matchers.h>
using qt_mocks::equals;
#include <qt-mocks/Verify.hpp>
// using QVERIFY_THAT

#include <map>
using std::map;

#include <accounting/AccountCode.h>
using drn::accounting::AccountCode;
#include <accounting/AccountingErrors.h>
using drn::accounting::AccountError;
#include <accounting/AccountNumber.h>
using drn::accounting::AccountNumber;
#include <banking/BankAccount.h>
using drn::banking::BankAccount;
#include <banking/BankAccountType.h>
using drn::banking::BankAccountType;
using drn::banking::SupportedAccountTypes;
#include <banking/BankingErrors.h>
using drn::banking::BankError;
#include <banking/BankName.h>
using drn::banking::BankName;
#include <banking/Reconciliation.h>
using drn::banking::ReconciledBankAccount;
#include <budgeting/BudgetingErrors.h>
using drn::budgeting::BudgetItemError;
#include <budgeting/BudgetItemIdentifier.h>
using drn::budgeting::BudgetItemIdentifier;
#include <budgeting/BudgetItemTypes.h>
using drn::budgeting::BudgetItemTypes;
#include <budgeting/BudgetSource.h>
using drn::budgeting::BudgetSource;
#include <foundation/Error.h>
using drn::foundation::Error;
#include <foundation/Optional.hpp>
using drn::foundation::inPlace;
using drn::foundation::Optional;
#include <surveying/ReconciledSurvey.h>
using drn::surveying::ReconciledSurvey;
using drn::surveying::ReconciledSurveys;


namespace drn
{
namespace surveying
{
namespace tests
{

struct ReconciledSurveysUT : QObject
{
	Q_OBJECT

	const map<BudgetItemIdentifier, Money> distribution_{
		{
			{
				BudgetItemIdentifier{BudgetItemTypes::Bill, BudgetSource{"Web Hosting"}},
				Money{50, 1234, Iso4217Codes::USD}
			}
		}
	};
	const ReconciledBankAccount reconciled_{
		BankAccount{
			BankName{"C1st Credit Union"},
			BankAccountType{
				AccountCode{AccountNumber{4321}, "Primary Share"},
				SupportedAccountTypes::Savings
			}
		},
		Money{50, 1234, Iso4217Codes::USD},
		Optional<QDate>{inPlace, 2022, 10, 3}
	};
	const Optional<QString> checksum_{
		inPlace,
		QStringLiteral("JCgKc2N7TCsONzuTKBmeEZxz9rwRH7DLBetXA/HcaSs=")
	};
	const ReconciledSurvey survey_{this->reconciled_, this->distribution_, this->checksum_};

private slots:
	//{ constructor Tests

	void constructor_default_ShouldBeDefaults()
	{
		const ReconciledSurveys surveys{};
		QVERIFY_TRUE(surveys.empty());

	}

	void constructor_fullInit_ShouldBeValues()
	{
		const ReconciledSurveys surveys{{this->reconciled_.bankAccount(), this->survey_}};
		QVERIFY_THAT(surveys.size(), equals(static_cast<size_t>(1)));
		const auto survey{*surveys.cbegin()};
		QVERIFY_THAT(survey.first, equals(this->reconciled_.bankAccount()));
		QVERIFY_THAT(survey.second, equals(this->survey_));
	}

	//}

	//{ upsert Tests

	void upsert_newEntry_ShouldInsert()
	{
		ReconciledSurveys surveys{};
		QVERIFY_TRUE(surveys.upsert(this->survey_));
		QVERIFY_THAT(surveys.size(), equals(static_cast<size_t>(1)));
		const auto survey{*surveys.cbegin()};
		QVERIFY_THAT(survey.first, equals(this->reconciled_.bankAccount()));
		QVERIFY_THAT(survey.second, equals(this->survey_));
	}

	void upsert_existingEntry_ShouldUpdate()
	{
		ReconciledSurveys surveys{{this->reconciled_.bankAccount(), this->survey_}};
		const ReconciledSurvey newSurvey{
			ReconciledBankAccount{
				this->reconciled_.bankAccount(),
				this->reconciled_.balance(),
				Optional<QDate>{inPlace, 2022, 11, 2}
			},
			this->distribution_,
			Optional<QString>{"96G5Ye9Ts7uZaBoPoxauIOqr2xzIA9jrNCISvXCLJlY="}
		};
		QVERIFY_FALSE(surveys.upsert(newSurvey));
		QVERIFY_THAT(surveys.size(), equals(static_cast<size_t>(1)));
		const auto survey{*surveys.cbegin()};
		QVERIFY_THAT(survey.first, equals(this->reconciled_.bankAccount()));
		QVERIFY_THAT(survey.second, equals(newSurvey));
	}

	void upsert_entryHasItemInDistributionAlreadyDistributed_ShouldThrow()
	{
		ReconciledSurveys surveys{{this->reconciled_.bankAccount(), this->survey_}};
		const ReconciledSurvey newSurvey{
			ReconciledBankAccount{
				BankAccount{
					BankName{"C1st Credit Union"},
					BankAccountType{
						AccountCode{AccountNumber{9876}, "Free"},
						SupportedAccountTypes::Chequeing
					}
				},
				this->reconciled_.balance(),
				Optional<QDate>{inPlace, 2022, 11, 2}
			},
			this->distribution_,
			Optional<QString>{"HY8X2Zu6erlP02jo4JWtkyrQdTdK7jBcpBvzmjdlDto="}
		};
		QVERIFY_EXCEPTION_THROWN(surveys.upsert(newSurvey), BankError);
		QVERIFY_THAT(surveys.size(), equals(static_cast<size_t>(1)));
		const auto survey{*surveys.cbegin()};
		QVERIFY_THAT(survey.first, equals(this->reconciled_.bankAccount()));
		QVERIFY_THAT(survey.second, equals(this->survey_));
	}

	//}
};


}}}

QTEST_MAIN(drn::surveying::tests::ReconciledSurveysUT)
#include "ReconciledSurveys-unit-tests.moc"
